@@ -1,10 +1,11 @@
 use crate::{
-    mutex::{Arc, Mutex, RwLock},
+    mutex::{Mutex, RwLock},
     TextureAtlas,
 };
 use ahash::AHashMap;
 use emath::{vec2, Vec2};
 use std::collections::BTreeSet;
+use std::sync::Arc;
 
 // ----------------------------------------------------------------------------
 
@@ -212,7 +213,7 @@ impl FontImpl {
 type FontIndex = usize;
 
 // TODO: rename?
-/// Wrapper over multiple `FontImpl` (e.g. a primary + fallbacks for emojis)
+/// Wrapper over multiple [`FontImpl`] (e.g. a primary + fallbacks for emojis)
 pub struct Font {
     fonts: Vec<Arc<FontImpl>>,
     /// Lazily calculated.
@@ -377,7 +378,7 @@ fn allocate_glyph(
                 if v > 0.0 {
                     let px = glyph_pos.0 + x as usize;
                     let py = glyph_pos.1 + y as usize;
-                    image[(px, py)] = fast_round(v * 255.0);
+                    image[(px, py)] = v;
                 }
             });
 
@@ -405,7 +406,3 @@ fn allocate_glyph(
         uv_rect,
     }
 }
-
-fn fast_round(r: f32) -> u8 {
-    (r + 0.5).floor() as _ // rust does a saturating cast since 1.45
-}
